#[macro_use]
extern crate rocket;

#[macro_use]
extern crate rocket_contrib;

// Seemingly necessary to get serde::Serialize into scope for Prost
// code generation.
#[macro_use]
extern crate serde_derive;

use rocket_contrib::serve::StaticFiles;
use rocket_contrib::templates::Template;
use std::env;

pub mod catchers;
pub mod db;
pub mod errors;
pub mod migrator;
pub mod models;
pub mod routes;

async fn gin_and_tonic() -> Result<(), Box<dyn std::error::Error>> {
    use crate::models::proto::cofd::api::cofd_api_server::{CofdApi, CofdApiServer};
    use crate::models::proto::cofd::api::UpdateSkillValueRequest;
    use crate::models::proto::cofd::cofd_sheet::Skill;
    use tonic::{transport::Server, Request, Response, Status};

    #[derive(Debug, Default)]
    pub struct MyGreeter {}

    #[tonic::async_trait]
    impl CofdApi for MyGreeter {
        async fn update_skill_value(
            &self,
            request: Request<UpdateSkillValueRequest>, // Accept request of type HelloRequest
        ) -> Result<Response<Skill>, Status> {
            // Return an instance of type HelloReply
            println!("Got a request: {:?}", request);

            let reply = Skill::default();

            Ok(Response::new(reply)) // Send back our formatted greeting
        }
    }

    let addr = "[::1]:50051".parse()?;
    let greeter = MyGreeter::default();

    println!("Running tonic");
    Server::builder()
        .add_service(CofdApiServer::new(greeter))
        .serve(addr)
        .await?;

    println!("ending tonic");
    Ok(())
}

async fn make_rocket(database: sqlx::SqlitePool) -> Result<(), Box<dyn std::error::Error>> {
    println!("Running rocket");
    let root_routes: Vec<rocket::Route> = {
        routes::root::routes()
            .into_iter()
            .chain(routes::auth::routes().into_iter())
            .collect()
    };

    let api_routes = routes::api::routes();
    let character_routes = routes::characters::routes();
    let catchers = catchers::catchers();

    rocket::ignite()
        .attach(Template::fairing())
        //.attach(db::TenebrousDbConn::fairing())
        .manage(database)
        .mount("/", root_routes)
        .mount("/characters", character_routes)
        .mount("/api", api_routes)
        .mount(
            "/scripts",
            StaticFiles::from(concat!(env!("CARGO_MANIFEST_DIR"), "/static/scripts")),
        )
        .mount(
            "/protos",
            StaticFiles::from(concat!(env!("CARGO_MANIFEST_DIR"), "/proto")),
        )
        .register(catchers)
        .launch()
        .await?;

    Ok(())
}

#[rocket::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let args: Vec<String> = env::args().collect();
    let db_path: &str = match &args[..] {
        [_, path] => path.as_ref(),
        [_, _, ..] => panic!("Expected exactly 0 or 1 argument"),
        _ => "tenebrous.sqlite",
    };

    println!("Using database: {}", db_path);

    migrator::migrate(db_path).await?;

    let db = crate::db::create_pool(db_path).await?;
    // let tonic = gin_and_tonic();
    // let rocket = make_rocket(db);

    tokio::select! {
        result = make_rocket(db) => {
            println!("done with rocket: {:?}", result);
        }
        result = gin_and_tonic() => {
            println!("done with tonic: {:?}", result);
        }
    }

    Ok(())
}
